signature PRINT =
sig
  (* print an expression *)
  val expToString : Syntax.exp -> string
  val typToString : Syntax.typ -> string
end

structure Print :> PRINT =
struct

  open Syntax

  fun nameToString (x,s) = x ^ "[" ^ Int.toString s ^"]"
      
  fun typ t = 
      case t of 
          Nat => "nat"
        | Arrow(t1,t2) => "( " ^ typ t1 ^ " -> " ^ typ t2 ^ " )"

  fun exp0 (tab, e) = case e
       of App(e1, e2) =>
              exp0 (tab, e1) ^ "@" ^ expinf (tab, e2)
	| _ => expinf (tab, e)

  and expinf (tab, e) = case e
      of Zero => "z"
    | Succ e => (case detectNumeral e of
                  NONE => "s "^(exp0 (tab,e))
                | SOME n => Int.toString (n+1))
    | Var(x) => nameToString x
    | Lam(t,(x,e)) => "(lam " ^ nameToString x ^ " : " ^ typ t ^ "." ^ exp0 (tab,e) ^")"
    | Rec (e0,(x,y,e1),e) => 
          "rec " ^ exp0 (tab,e) ^ " { z => " ^ exp0 (tab,e0) ^ " | s " 
              ^ nameToString x ^ " with " ^ nameToString y ^ " => " ^ exp0 (tab,e1) ^ "}"
    | Let(e1, (x, e2)) =>
          let val tab' = tab ^ "    "
          in "let " ^ nameToString x ^ " be " ^ exp0 (tab, e1) ^ " in" ^
              tab' ^ exp0 (tab', e2) ^
              tab ^ "end" end
    | _ => "(" ^ exp0 (tab, e) ^ ")"

  and detectNumeral e = case e of 
      Zero => SOME 0
    | Succ e' => (case detectNumeral e' of
                   NONE => NONE
                 | SOME n' => SOME (n' + 1))
    | _ => NONE
          
      
              
  fun expToString t = exp0 ("\n", t)
  fun typToString t = typ t

end;  (* functor PrintFun *)
